#include "vtkRenderer.h"
#include "vtkRenderWindow.h"
#include "vtkRenderWindowInteractor.h"
#include "vtkPolyData.h"
#include "vtkPolyDataMapper.h"
#include "vtkPolyDataWriter.h"
#include "vtkDataSetMapper.h"
#include "vtkPoints.h"
#include "vtkCellArray.h"
#include "vtkIdList.h"
#include "vtkActor.h"
#include "vtkProperty.h"
#include "vtkCamera.h"

#include "vtkPolyData.h"
#include "vtkPolyDataWriter.h"
#include "vtkXMLPolyDataReader.h"

#include "vtkTimerLog.h"
#include "vtkDoubleArray.h"
#include "vtkMath.h"

#include "vtkCommand.h"

#include "vtkRBFDeformation.h"


class vtkTimerCallback : public vtkCommand
{
public:
  static vtkTimerCallback *New()
  {
    vtkTimerCallback *cb = new vtkTimerCallback;
    cb->FastTimerId = 0;
    cb->FasterTimerId = 0;
    cb->RenderTimerId = 0;
    return cb;
  }

  virtual void Execute(vtkObject *caller, unsigned long eventId, void *callData)
  {
    if (vtkCommand::TimerEvent == eventId)
    {
      int tid = * static_cast<int *>(callData);

      if (tid == this->FastTimerId)
      {
        int id = vtkMath::Round(10*vtkMath::Random());
        double d[3];
        d[0] = 0.1; d[1] = 0.2;d[2] = 0.4;
        this->RBF->SetDisplacement(id,d);
      }
      else if (tid == this->FasterTimerId)
      {
        vtkTimerLog * timer = vtkTimerLog::New();
        timer->StartTimer();
        this->RBF->Modified();
        this->RBF->Update();
        timer->StopTimer();

        //std::cout << "[Test] Execution Rate: " << 1/(timer->GetElapsedTime()) << "\n";

      }
      else if (tid == this->RenderTimerId)
      {
        vtkRenderWindowInteractor *iren = vtkRenderWindowInteractor::SafeDownCast(caller);
        if (iren && iren->GetRenderWindow() && iren->GetRenderWindow()->GetRenderers())
        {
          iren->Render();
        }
      }
    }
  }

  void SetFastTimerId(int tid)
  {
    this->FastTimerId = tid;
  }

  void SetFasterTimerId(int tid)
  {
    this->FasterTimerId = tid;
  }

  void SetRenderTimerId(int tid)
  {
    this->RenderTimerId = tid;
  }

  void SetRBF(vtkRBFDeformation * RBF)
  {
    this->RBF = RBF;
  }

  void SetContactIds(vtkIdList * list)
  {
    this->List = list;
  }
private:
  int FastTimerId;
  int RenderTimerId;
  int FasterTimerId;

  vtkIdList * List;

  vtkRBFDeformation * RBF;
};

int main(int argc, char * argv[])
{
  const char * filename = "/home/jballesteros/Workspace/data/vtkESQuiData/Scenario/Meshes/sphere12_12_1.vtp";

  if (argc > 1)
  {
    filename = argv[1];
  }

  vtkXMLPolyDataReader * reader = vtkXMLPolyDataReader::New();
  reader->SetFileName(filename);
  reader->Update();

  vtkPolyData * mesh = reader->GetOutput();

  std::cout << "[Test] Input grid #points: " << mesh->GetNumberOfPoints() << "\n";
  std::cout << "[Test] Input grid #cells: " << mesh->GetNumberOfCells() << "\n";

  //Float to Double Conversion
  if(mesh->GetPoints()->GetDataType() != VTK_DOUBLE)
  {
    vtkPoints * ps = vtkPoints::New();
    ps->SetDataTypeToDouble();
    ps->DeepCopy(mesh->GetPoints());
    mesh->SetPoints(ps);
    cout << "DataType has been modified\n";
  }

  mesh->Update();

  vtkRBFDeformation* rbf = vtkRBFDeformation::New();
  rbf->SetInput(mesh);
  rbf->SetNumberOfControlPoints(15);
  rbf->SetDeltaT(0.1);
  rbf->SetRelaxationFactor(0.001);
  rbf->Init();

  vtkRenderer * renderer = vtkRenderer::New();
  vtkRenderWindow * renWin = vtkRenderWindow::New();
  renWin->SetSize(500,500);
  renWin->AddRenderer(renderer);

  vtkRenderWindowInteractor * iren = vtkRenderWindowInteractor::New();
  iren->SetRenderWindow(renWin);

  vtkDataSetMapper * mapper = vtkDataSetMapper::New();
  mapper->SetInput(mesh);
  mapper->ScalarVisibilityOff();

  vtkActor * actor = vtkActor::New();
  actor->SetMapper(mapper);
  actor->GetProperty()->SetColor(0,1,0);
  actor->GetProperty()->SetOpacity(0.5);

  vtkDataSetMapper * mapper2 = vtkDataSetMapper::New();
  mapper2->SetInput(rbf->GetOutput());
  mapper2->ScalarVisibilityOff();

  vtkActor * actor2 = vtkActor::New();
  actor2->SetMapper(mapper2);
  actor2->GetProperty()->SetColor(1,0,0);
  actor2->GetProperty()->SetRepresentationToWireframe();

  renderer->AddActor(actor);
  renderer->AddActor(actor2);
  renderer->SetBackground(1,1,1);

  renderer->ResetCamera();
  renderer->GetActiveCamera()->Azimuth(90);
  iren->Initialize();

  renWin->Render();

  vtkMath::RandomSeed(1);

  // Sign up to receive TimerEvent:
  //
  vtkTimerCallback *cb = vtkTimerCallback::New();
  iren->AddObserver(vtkCommand::TimerEvent, cb);
  int tid;

  cb->SetRBF(rbf);

  //Create a fast timer for displacement
  tid = iren->CreateRepeatingTimer(2000);
  cb->SetFastTimerId(tid);

  //Create a faster timer for update
  tid = iren->CreateRepeatingTimer(1);
  cb->SetFasterTimerId(tid);

  // Create a slower repeating timer to trigger Render calls.
  // (This fires at the rate of approximately 25 frames per second.)
  //
  tid = iren->CreateRepeatingTimer(40);
  cb->SetRenderTimerId(tid);

  iren->Start();

  rbf->Delete();
  mapper->Delete();
  actor->Delete();
  mapper2->Delete();
  actor2->Delete();
  renderer->Delete();
  renWin->Delete();
  iren->Delete();
  return 0;
}

